#include <stdio.h>
#include <math.h>		/* for floor() */


void decode_date(int date, int *year, int *month, int *day)
{
  int idate;

  *year  =  date / 10000;
  idate  = date - *year*10000;
  if ( idate < 0 ) idate = -idate;
  *month = idate / 100;
  *day   = idate - *month*100;
}


int encode_date(int year, int month, int day)
{
  int date;
  int iyear;

  iyear = year;
  if ( iyear < 0 ) iyear = -iyear;
  date = iyear*10000 + month*100 + day;
  if ( year < 0 ) date = -date;

  return (date);
}


void decode_time(int time, int *hour, int *minute)
{
  *hour   =  time / 100;
  *minute =  time - *hour*100;
}


int encode_time(int hour, int minute)
{
  int time;

  time = hour*100 + minute;

  return (time);
}


/* convert Julian date into year, months, day */
void decode_julday(
    int  	       julday,		/* Julian day number to convert */
    int			*year,		/* Gregorian year (out) */
    int			*mon,		/* Gregorian month (1-12) (out) */
    int			*day		/* Gregorian day (1-31) (out) */
    )
{
  int a = julday;
  double b, c;
  double d, e, f;

  if ( a < 2299161 )
    {
      c = a + 1524;
    } 
  else
    {
      b = floor((a - 1867216.25)/36524.25);
      c = a + b - floor(b/4) + 1525;
    }

  d = floor((c - 122.1)/365.25);
  e = floor(365.25*d);
  f = floor((c - e)/30.6001);

  *day  = (int)(c - e - floor(30.6001*f));
  *mon  = (int)(f - 1 - 12*floor(f/14));
  *year = (int)(d - 4715 - floor((7 + *mon)/10));
}

/* convert year, month, day into Julian calendar day */
int encode_julday(int year, int month, int day)
{
  int iy;
  int im;
  int ib;
  int julday;

  if ( month <= 2 )
    {
      iy = year  - 1;
      im = month + 12;
    }
  else
    {
      iy = year;
      im = month;
    }


  if ( iy < 0 )
    ib = (int)((iy+1)/400) - (int)((iy+1)/100);
  else
    ib = (int)(iy/400) - (int)(iy/100);

  if ( year > 1582 || (year == 1582 && (month > 10 || (month == 10 && day >= 15))) )
    {
      /*
      ** 15th October 1582 AD or later
      */
    }
  else
    {
      /*
      ** 4th October 1582 AD or earlier
      */
      ib = -2;
    }

  julday = (int) (floor(365.25*iy) + (int)(30.6001*(im+1)) + ib + 1720996.5 + day + 0.5);

  return (julday);
}


int date_to_julday(int date)
{
  int julday;
  int year, month, day;

  decode_date(date, &year, &month, &day);

  julday = encode_julday(year, month, day);

  return (julday);
}


int julday_to_date(int julday)
{
  int date;
  int year, month, day;

  decode_julday(julday, &year, &month, &day);

  date = encode_date(year, month, day);

  return (date);
}


int time_to_sec(int time)
{
  int secofday;
  int hour, minute;

  decode_time(time, &hour, &minute);

  secofday = (hour*60 + minute)*60;

  return (secofday);
}


int sec_to_time(int secofday)
{
  int time;
  int hour, minute;

  hour   = secofday/3600;
  minute = secofday/60 - hour*60;

  time = hour*100 + minute;

  return (time);
}


void julday_add_seconds(int seconds, int *julday, int *secofday)
{
  int days;

  *secofday += seconds;

  days = *secofday/86400;

  *secofday -= days*86400;

  *julday += days;
}

/* add days and secs to julday/secofday */
void julday_add(int days, int secs, int *julday, int *secofday)
{
  int mdays;

  *julday += days;
  *secofday += secs;

  mdays = *secofday/86400;

  if ( mdays > 0 )
    {
      *julday += mdays;
      *secofday -= mdays*86400;
    }
}

/* subtract julday1/secofday1 from julday2/secofday2 and returns the result in seconds */
int julday_sub(int julday1, int secofday1, int julday2, int secofday2, int *days, int *secs)
{
  int seconds;

  *days = julday2 - julday1;
  *secs = secofday2 - secofday1;

  if ( *secs < 0 )
    {
      *days -= 1;
      *secs += 86400;
    }

  seconds = *days*86400 + *secs;

  return (seconds);
}


void encode_juldaysec(int year, int month, int day, int hour, int minute, int *julday, int *secofday)
{
  *julday = encode_julday(year, month, day);

  *secofday = (hour*60 + minute)*60;
}


void decode_juldaysec(int julday, int secofday, int *year, int *month, int *day, int *hour, int *minute)
{
  decode_julday(julday, year, month, day);

  *hour   = secofday/3600;
  *minute = secofday/60 - *hour*60;
}


#ifdef TEST
int main(void)
{
  int nmin;
  int vdate0, vtime0;
  int vdate, vtime;
  int ijulinc;
  int i, j = 0;
  int year, mon, day, hour, min;
  int julday, secofday;

  /* 1 - Check valid range of years */

  nmin = 11000;
  vdate0 = -80001201;
  vtime0 = 1205;

  printf("start time: %8d %4d\n", vdate0, vtime0);

  for ( i = 0; i < nmin; i++ )
    {
      decode_date(vdate0, &year, &mon, &day);
      decode_time(vtime0, &hour, &min);

      julday  = date_to_julday(vdate0);
      secofday = time_to_sec(vtime0);

      vdate = julday_to_date(julday);
      vtime = sec_to_time(secofday);

      if ( vdate0 != vdate || vtime0 != vtime )
	printf("%4d %8d %4d %8d %4d %9d %9d\n",
	       ++j, vdate0, vtime0, vdate, vtime, julday, secofday);

      year++;
      vdate0 = encode_date(year, mon, day);
      vtime0 = encode_time(hour, min);
    }

  printf("stop time: %8d %4d\n", vdate0, vtime0);

  /* 2 - Check time increment of one minute */

  nmin = 120000;
  ijulinc = 60;
  vdate0 = 20001201;
  vtime0 = 0;

  printf("start time: %8d %4d\n", vdate0, vtime0);

  julday = date_to_julday(vdate0);
  secofday = time_to_sec(vtime0);
  for ( i = 0; i < nmin; i++ )
    {
      decode_date(vdate0, &year, &mon, &day);
      decode_time(vtime0, &hour, &min);

      if ( ++min >= 60 )
	{
	  min = 0;
	  if ( ++hour >= 24 )
	    {
	      hour = 0;
	      if ( ++day >= 32 )
		{
		  day = 1;
		  if ( ++mon >= 13 )
		    {
		      mon = 1;
		      year++;
		    }
		}
	    }
	}

      vdate0 = encode_date(year, mon, day);
      vtime0 = encode_time(hour, min);

      julday_add_seconds(ijulinc, &julday, &secofday);

      vdate = julday_to_date(julday);
      vtime = sec_to_time(secofday);
      if ( vdate0 != vdate || vtime0 != vtime )
	printf("%4d %8d %4d %8d %4d %9d %9d\n",
	       ++j, vdate0, vtime0, vdate, vtime, julday, secofday);
    }

  printf("stop time: %8d %4d\n", vdate0, vtime0);

  return (0);
}
#endif


#ifdef TEST2
int main(void)
{
  int i;
  int julday, secofday;
  int year, month, day, hour, minute;
  int value = 30;
  int factor = 86400;

  year=1979; month=1; day=15; hour=12; minute=30;

  printf("%d/%02d/%02d %02d:%02d\n", year, month, day, hour, minute);

  encode_juldaysec(year, month, day, hour, minute, &julday, &secofday);

  decode_juldaysec(julday, secofday, &year, &month, &day, &hour, &minute);
  printf("%d/%02d/%02d %02d:%02d   %d %d\n", year, month, day, hour, minute, julday, secofday);

  for ( i = 0; i < 420; i++ )
    {

      decode_juldaysec(julday, secofday, &year, &month, &day, &hour, &minute);
      printf("%2d %d/%02d/%02d %02d:%02d\n", i, year, month, day, hour, minute);
      julday_add_seconds(value*factor, &julday, &secofday);
    }

  return (0);
}
#endif
